@@ -15,21 +15,21 @@
 # specific language governing permissions and limitations
 # under the License.
 
+import base64
 import logging
+import platform
 import socket
 import string
-import base64
+
+import urllib3
 
 try:
-    import http.client as httplib
-    from urllib import request as url_request
     from urllib import parse
 except ImportError:  # above is available in py3+, below is py2.7
-    import httplib as httplib
-    import urllib2 as url_request
     import urlparse as parse
 
 from selenium.webdriver.common import utils as common_utils
+from selenium import __version__
 from .command import Command
 from .errorhandler import ErrorCode
 from . import utils
@@ -37,98 +37,6 @@ from . import utils
 LOGGER = logging.getLogger(__name__)
 
 
-class Request(url_request.Request):
-    """
-    Extends the url_request.Request to support all HTTP request types.
-    """
-
-    def __init__(self, url, data=None, method=None):
-        """
-        Initialise a new HTTP request.
-
-        :Args:
-            - url - String for the URL to send the request to.
-            - data - Data to send with the request.
-        """
-        if method is None:
-            method = data is not None and 'POST' or 'GET'
-        elif method != 'POST' and method != 'PUT':
-            data = None
-        self._method = method
-        url_request.Request.__init__(self, url, data=data)
-
-    def get_method(self):
-        """
-        Returns the HTTP method used by this request.
-        """
-        return self._method
-
-
-class Response(object):
-    """
-    Represents an HTTP response.
-    """
-
-    def __init__(self, fp, code, headers, url):
-        """
-        Initialise a new Response.
-
-        :Args:
-            - fp - The response body file object.
-            - code - The HTTP status code returned by the server.
-            - headers - A dictionary of headers returned by the server.
-            - url - URL of the retrieved resource represented by this Response.
-        """
-        self.fp = fp
-        self.read = fp.read
-        self.code = code
-        self.headers = headers
-        self.url = url
-
-    def close(self):
-        """
-        Close the response body file object.
-        """
-        self.read = None
-        self.fp = None
-
-    def info(self):
-        """
-        Returns the response headers.
-        """
-        return self.headers
-
-    def geturl(self):
-        """
-        Returns the URL for the resource returned in this response.
-        """
-        return self.url
-
-
-class HttpErrorHandler(url_request.HTTPDefaultErrorHandler):
-    """
-    A custom HTTP error handler.
-
-    Used to return Response objects instead of raising an HTTPError exception.
-    """
-
-    def http_error_default(self, req, fp, code, msg, headers):
-        """
-        Default HTTP error handler.
-
-        :Args:
-            - req - The original Request object.
-            - fp - The response body file object.
-            - code - The HTTP status code returned by the server.
-            - msg - The HTTP status message returned by the server.
-            - headers - The response headers.
-
-        :Returns:
-            A new Response object.
-        """
-        return Response(fp, code, headers, req.get_full_url())
-
-
 class RemoteConnection(object):
     """A connection with the Remote WebDriver server.
 
@@ -162,18 +70,56 @@ class RemoteConnection(object):
         """
         cls._timeout = socket._GLOBAL_DEFAULT_TIMEOUT
 
+    @classmethod
+    def get_remote_connection_headers(cls, parsed_url, keep_alive=False):
+        """
+        Get headers for remote request.
+
+        :Args:
+         - parsed_url - The parsed url
+         - keep_alive (Boolean) - Is this a keep-alive connection (default: False)
+        """
+
+        system = platform.system().lower()
+        if system == "darwin":
+            system = "mac"
+
+        headers = {
+            'Accept': 'application/json',
+            'Content-Type': 'application/json;charset=UTF-8',
+            'User-Agent': 'selenium/{} (python {})'.format(__version__, system)
+        }
+
+        if parsed_url.username:
+            base64string = base64.b64encode('{0.username}:{0.password}'.format(parsed_url).encode())
+            headers.update({
+                'Authorization': 'Basic {}'.format(base64string.decode())
+            })
+
+        if keep_alive:
+            headers.update({
+                'Connection': 'keep-alive'
+            })
+
+        return headers
+
     def __init__(self, remote_server_addr, keep_alive=False, resolve_ip=True):
         # Attempt to resolve the hostname and get an IP address.
         self.keep_alive = keep_alive
         parsed_url = parse.urlparse(remote_server_addr)
-        addr = parsed_url.hostname
         if parsed_url.hostname and resolve_ip:
             port = parsed_url.port or None
-            ip = common_utils.find_connectable_ip(parsed_url.hostname,
-                                                  port=port)
+            if parsed_url.scheme == "https":
+                ip = parsed_url.hostname
+            elif port and not common_utils.is_connectable(port, parsed_url.hostname):
+                ip = None
+                LOGGER.info('Could not connect to port {} on host '
+                            '{}'.format(port, parsed_url.hostname))
+            else:
+                ip = common_utils.find_connectable_ip(parsed_url.hostname,
+                                                      port=port)
             if ip:
                 netloc = ip
-                addr = netloc
                 if parsed_url.port:
                     netloc = common_utils.join_host_port(netloc,
                                                          parsed_url.port)
@@ -191,8 +137,7 @@ class RemoteConnection(object):
 
         self._url = remote_server_addr
         if keep_alive:
-            self._conn = httplib.HTTPConnection(
-                str(addr), str(parsed_url.port), timeout=self._timeout)
+            self._conn = urllib3.PoolManager(timeout=self._timeout)
 
         self._commands = {
             Command.STATUS: ('GET', '/status'),
@@ -201,13 +146,21 @@ class RemoteConnection(object):
             Command.QUIT: ('DELETE', '/session/$sessionId'),
             Command.GET_CURRENT_WINDOW_HANDLE:
                 ('GET', '/session/$sessionId/window_handle'),
+            Command.W3C_GET_CURRENT_WINDOW_HANDLE:
+                ('GET', '/session/$sessionId/window'),
             Command.GET_WINDOW_HANDLES:
                 ('GET', '/session/$sessionId/window_handles'),
+            Command.W3C_GET_WINDOW_HANDLES:
+                ('GET', '/session/$sessionId/window/handles'),
             Command.GET: ('POST', '/session/$sessionId/url'),
             Command.GO_FORWARD: ('POST', '/session/$sessionId/forward'),
             Command.GO_BACK: ('POST', '/session/$sessionId/back'),
             Command.REFRESH: ('POST', '/session/$sessionId/refresh'),
             Command.EXECUTE_SCRIPT: ('POST', '/session/$sessionId/execute'),
+            Command.W3C_EXECUTE_SCRIPT:
+                ('POST', '/session/$sessionId/execute/sync'),
+            Command.W3C_EXECUTE_SCRIPT_ASYNC:
+                ('POST', '/session/$sessionId/execute/async'),
             Command.GET_CURRENT_URL: ('GET', '/session/$sessionId/url'),
             Command.GET_TITLE: ('GET', '/session/$sessionId/title'),
             Command.GET_PAGE_SOURCE: ('GET', '/session/$sessionId/source'),
@@ -255,10 +208,9 @@ class RemoteConnection(object):
                 ('GET', '/session/$sessionId/element/$id/attribute/$name'),
             Command.GET_ELEMENT_PROPERTY:
                 ('GET', '/session/$sessionId/element/$id/property/$name'),
-            Command.ELEMENT_EQUALS:
-                ('GET', '/session/$sessionId/element/$id/equals/$other'),
             Command.GET_ALL_COOKIES: ('GET', '/session/$sessionId/cookie'),
             Command.ADD_COOKIE: ('POST', '/session/$sessionId/cookie'),
+            Command.GET_COOKIE: ('GET', '/session/$sessionId/cookie/$name'),
             Command.DELETE_ALL_COOKIES:
                 ('DELETE', '/session/$sessionId/cookie'),
             Command.DELETE_COOKIE:
@@ -278,16 +230,28 @@ class RemoteConnection(object):
                 ('POST', '/session/$sessionId/timeouts'),
             Command.DISMISS_ALERT:
                 ('POST', '/session/$sessionId/dismiss_alert'),
+            Command.W3C_DISMISS_ALERT:
+                ('POST', '/session/$sessionId/alert/dismiss'),
             Command.ACCEPT_ALERT:
                 ('POST', '/session/$sessionId/accept_alert'),
+            Command.W3C_ACCEPT_ALERT:
+                ('POST', '/session/$sessionId/alert/accept'),
             Command.SET_ALERT_VALUE:
                 ('POST', '/session/$sessionId/alert_text'),
+            Command.W3C_SET_ALERT_VALUE:
+                ('POST', '/session/$sessionId/alert/text'),
             Command.GET_ALERT_TEXT:
                 ('GET', '/session/$sessionId/alert_text'),
+            Command.W3C_GET_ALERT_TEXT:
+                ('GET', '/session/$sessionId/alert/text'),
             Command.SET_ALERT_CREDENTIALS:
                 ('POST', '/session/$sessionId/alert/credentials'),
             Command.CLICK:
                 ('POST', '/session/$sessionId/click'),
+            Command.W3C_ACTIONS:
+                ('POST', '/session/$sessionId/actions'),
+            Command.W3C_CLEAR_ACTIONS:
+                ('DELETE', '/session/$sessionId/actions'),
             Command.DOUBLE_CLICK:
                 ('POST', '/session/$sessionId/doubleclick'),
             Command.MOUSE_DOWN:
@@ -298,20 +262,16 @@ class RemoteConnection(object):
                 ('POST', '/session/$sessionId/moveto'),
             Command.GET_WINDOW_SIZE:
                 ('GET', '/session/$sessionId/window/$windowHandle/size'),
-            Command.W3C_GET_WINDOW_SIZE:
-                ('GET', '/session/$sessionId/window/size'),
             Command.SET_WINDOW_SIZE:
                 ('POST', '/session/$sessionId/window/$windowHandle/size'),
-            Command.W3C_SET_WINDOW_SIZE:
-                ('POST', '/session/$sessionId/window/size'),
             Command.GET_WINDOW_POSITION:
                 ('GET', '/session/$sessionId/window/$windowHandle/position'),
             Command.SET_WINDOW_POSITION:
                 ('POST', '/session/$sessionId/window/$windowHandle/position'),
-            Command.W3C_GET_WINDOW_POSITION:
-                ('GET', '/session/$sessionId/window/position'),
-            Command.W3C_SET_WINDOW_POSITION:
-                ('POST', '/session/$sessionId/window/position'),
+            Command.SET_WINDOW_RECT:
+                ('POST', '/session/$sessionId/window/rect'),
+            Command.GET_WINDOW_RECT:
+                ('GET', '/session/$sessionId/window/rect'),
             Command.MAXIMIZE_WINDOW:
                 ('POST', '/session/$sessionId/window/$windowHandle/maximize'),
             Command.W3C_MAXIMIZE_WINDOW:
@@ -386,6 +346,10 @@ class RemoteConnection(object):
                 ('GET', '/session/$sessionId/contexts'),
             Command.SWITCH_TO_CONTEXT:
                 ('POST', '/session/$sessionId/context'),
+            Command.FULLSCREEN_WINDOW:
+                ('POST', '/session/$sessionId/window/fullscreen'),
+            Command.MINIMIZE_WINDOW:
+                ('POST', '/session/$sessionId/window/minimize')
         }
 
     def execute(self, command, params):
@@ -402,8 +366,10 @@ class RemoteConnection(object):
         """
         command_info = self._commands[command]
         assert command_info is not None, 'Unrecognised command %s' % command
-        data = utils.dump_json(params)
         path = string.Template(command_info[1]).substitute(params)
+        if hasattr(self, 'w3c') and self.w3c and isinstance(params, dict) and 'sessionId' in params:
+            del params['sessionId']
+        data = utils.dump_json(params)
         url = '%s%s' % (self._url, path)
         return self._request(command_info[0], url, body=data)
 
@@ -422,96 +388,53 @@ class RemoteConnection(object):
         LOGGER.debug('%s %s %s' % (method, url, body))
 
         parsed_url = parse.urlparse(url)
+        headers = self.get_remote_connection_headers(parsed_url, self.keep_alive)
+        resp = None
+        if body and method != 'POST' and method != 'PUT':
+            body = None
 
         if self.keep_alive:
-            headers = {"Connection": 'keep-alive', method: parsed_url.path,
-                       "User-Agent": "Python http auth",
-                       "Content-type": "application/json;charset=\"UTF-8\"",
-                       "Accept": "application/json"}
-            if parsed_url.username:
-                auth = base64.standard_b64encode(('%s:%s' % (
-                    parsed_url.username,
-                    parsed_url.password)).encode('ascii')).decode('ascii').replace('\n', '')
-                headers["Authorization"] = "Basic %s" % auth
-            if body and method != 'POST' and method != 'PUT':
-                body = None
-            try:
-                self._conn.request(method, parsed_url.path, body, headers)
-                resp = self._conn.getresponse()
-            except (httplib.HTTPException, socket.error):
-                self._conn.close()
-                raise
+            resp = self._conn.request(method, url, body=body, headers=headers)
 
             statuscode = resp.status
         else:
-            password_manager = None
-            if parsed_url.username:
-                netloc = parsed_url.hostname
-                if parsed_url.port:
-                    netloc += ":%s" % parsed_url.port
-                cleaned_url = parse.urlunparse((
-                    parsed_url.scheme,
-                    netloc,
-                    parsed_url.path,
-                    parsed_url.params,
-                    parsed_url.query,
-                    parsed_url.fragment))
-                password_manager = url_request.HTTPPasswordMgrWithDefaultRealm()
-                password_manager.add_password(None,
-                                              "%s://%s" % (parsed_url.scheme, netloc),
-                                              parsed_url.username,
-                                              parsed_url.password)
-                request = Request(cleaned_url, data=body.encode('utf-8'), method=method)
-            else:
-                request = Request(url, data=body.encode('utf-8'), method=method)
+            http = urllib3.PoolManager(timeout=self._timeout)
+            resp = http.request(method, url, body=body, headers=headers)
 
-            request.add_header('Accept', 'application/json')
-            request.add_header('Content-Type', 'application/json;charset=UTF-8')
-
-            if password_manager:
-                opener = url_request.build_opener(url_request.HTTPRedirectHandler(),
-                                                  HttpErrorHandler(),
-                                                  url_request.HTTPBasicAuthHandler(password_manager))
-            else:
-                opener = url_request.build_opener(url_request.HTTPRedirectHandler(),
-                                                  HttpErrorHandler())
-            resp = opener.open(request, timeout=self._timeout)
-            statuscode = resp.code
+            statuscode = resp.status
             if not hasattr(resp, 'getheader'):
                 if hasattr(resp.headers, 'getheader'):
                     resp.getheader = lambda x: resp.headers.getheader(x)
                 elif hasattr(resp.headers, 'get'):
                     resp.getheader = lambda x: resp.headers.get(x)
 
-        data = resp.read()
+        data = resp.data.decode('UTF-8')
         try:
             if 300 <= statuscode < 304:
                 return self._request('GET', resp.getheader('location'))
-            body = data.decode('utf-8').replace('\x00', '').strip()
             if 399 < statuscode <= 500:
-                return {'status': statuscode, 'value': body}
+                return {'status': statuscode, 'value': data}
             content_type = []
             if resp.getheader('Content-Type') is not None:
                 content_type = resp.getheader('Content-Type').split(';')
             if not any([x.startswith('image/png') for x in content_type]):
+
                 try:
-                    data = utils.load_json(body.strip())
+                    data = utils.load_json(data.strip())
                 except ValueError:
                     if 199 < statuscode < 300:
                         status = ErrorCode.SUCCESS
                     else:
                         status = ErrorCode.UNKNOWN_ERROR
-                    return {'status': status, 'value': body.strip()}
+                    return {'status': status, 'value': data.strip()}
 
-                assert type(data) is dict, (
-                    'Invalid server response body: %s' % body)
                 # Some of the drivers incorrectly return a response
                 # with no 'value' field when they should return null.
                 if 'value' not in data:
                     data['value'] = None
                 return data
             else:
-                data = {'status': 0, 'value': body.strip()}
+                data = {'status': 0, 'value': data}
                 return data
         finally:
             LOGGER.debug("Finished Request")
